# -*- coding: utf-8 -*-
# Author: tigflanker
# Date: 12 Dec 2018
# 数据探索

# 0. 导包/宏参设定
from matplotlib.backends.backend_pdf import PdfPages
import matplotlib.pyplot as plt
import matplotlib as mpl
import numpy as np
import pandas as pd

# 1. 数据探索
# 返回：
# Variable_pre_exp：数据探索列表
# obs_miss_rate：数据集每条观测的缺失率（横向）

def Data_Explore(datain  # 数据入
                 ,qtl=[0.25, 0.5, 0.75]  # 计算分位数情况 
                 ,ext_t=3  # 按照 Mean ± t * std方式计数极端值
                 ,samples=5  # 携带实际观测值 
                 ,plt_out=''  # 定义变量分布、相关描述图地址，若不定义输出路径则不输出
                 ,plt_show=False  # 是否在output中输出分布图
                 ,char_cate_threshold=10  # 定义绘制字符型（分类型）饼图时，最大类数
                 ,sps=4  # 频数分布图按4宫格or9宫格展示（9宫格参数尚未调好、勿使用）
                 ,calc_obs_na=False  # 计算观测缺失率
                 ,ecd='utf-8'
                 ):
                 
    '''
    # 参数说明
    datain                 : 数据输入（必须）
    qtl=[0.25, 0.5, 0.75]  : 计算分位数情况 
    ext_t=3                : 按照 Mean ± t * std方式计数极端值
    samples=5              : 携带实际观测值 
    plt_out=''             : 定义变量分布、相关描述图地址，若不定义输出路径则不输出
    plt_show=False         : 是否在output中输出分布图
    char_cate_threshold=10 : 定义绘制字符型（分类型）饼图时，最大类数
    sps=4                  : 频数分布图按4宫格or9宫格展示（9宫格参数尚未调好、勿使用）
    calc_obs_na=False      : 计算观测缺失率
    ecd='utf-8'            : 结果输出字符类型
    
    # Example:
    datain_test = pd.read_csv('http://biostat.mc.vanderbilt.edu/wiki/pub/Main/DataSets/titanic.txt')
    
    Variable_pre_exp, obs_miss_rate = Data_Explore(datain_test, plt_out='D:/Desktop/xxx')
    '''
    
    # 1.1 数据全局印象
    datatp = datain.dtypes
    char_cols = set(list(datatp[datatp == 'object'].index))
    nume_cols = set(list(datatp[datatp != 'object'].index))
    
    # 通用：缺失、类数
    # 数值：均值/标准差、最大/最小值、0值、分布
    # 字符：众数
    
    # 数值型或通用
    _N = datain.shape[0]
    _isnull_n = datain.isnull().sum().rename('na_n')
    _nunique = datain.nunique().rename('val_cate_n')
    
    _zero_n = (datain == 0).sum().rename('0_n')
    
    # 数据分布描述（此处将混合型字段中的字符值call missing） 
    _num_df = datain.apply(lambda x:pd.to_numeric(x, errors='coerce'))
    _desc = _num_df.describe(percentiles = qtl, include = 'all').T
    
    _desc['mean_std'] = _desc['mean'].apply(lambda x:str(round(x, 4))) + '|' \
                      + _desc['std'].apply(lambda x:str(round(x, 4)))
    _desc['min_max'] = _desc['min'].apply(lambda x:str(x)) + '|' \
                     + _desc['max'].apply(lambda x:str(x))
    _desc['skew_kurt'] = _num_df.skew().apply(lambda x:str(round(x, 4))) + '|' \
                       + _num_df.kurt().apply(lambda x:str(round(x, 4)))
    _pvar = 'p_'+'_'.join([str(int(x * 100)) for x in qtl])   
    _desc[_pvar] = eval("+'|'+".join(["_desc['"+str(int(x * 100))+
         "%'].apply(lambda x:str(round(x, 4)))" for x in qtl]))

    # 计算极端值延伸的计算：两界极端值占比、去极端值均数/标准差、最大最小值
    _ms_dt = _desc[['mean', 'std']].dropna().T.to_dict()  # 将均数、标准差做成字典，供极端值计算 
    
    def ext_calc(col):
        _en = _desc['count'][col]    
        
        _elown = np.sum(_num_df[col] < _ms_dt[col]['mean'] - ext_t * _ms_dt[col]['std'])
        _euppn = np.sum(_num_df[col] > _ms_dt[col]['mean'] + ext_t * _ms_dt[col]['std'])
    
        _edesc = _num_df[col][_num_df[col].between(_ms_dt[col]['mean'] - ext_t * _ms_dt[col]['std'], 
                         _ms_dt[col]['mean'] + ext_t * _ms_dt[col]['std'])].describe()
    
        return (str(round(_elown * 100 / _en, 2)) + '%|' + str(round(_euppn * 100 / _en, 2)) + '%'
                ,str(round(_edesc['mean'], 4)) + '|' + str(round(_edesc['std'], 4))
                ,str(_edesc['min']) + '|' + str(_edesc['max']))
                
    # 框架、众数、极端值延伸 
    def vpe(col):
        vpe = [col  # 字段名
               , 'Character' if col in char_cols else 'Numeric'  # 字段类型
               , datain[col].mode()[0] if _nunique[col] > 0 else np.nan  # 众数
               
               , ext_calc(col) if col in _ms_dt else ('nan|nan', 'nan|nan', 'nan|nan')  # 极端值相关，待拆包
               ]
        return vpe
    
    Variable_pre_exp = pd.DataFrame(list(map(vpe, datain.columns)), columns=['var', 'var_type', 'md_v', 'ext_mass'])
    Variable_pre_exp['ext_low_upp_r'] = Variable_pre_exp.ext_mass.apply(lambda x:x[0])
    Variable_pre_exp['noext_mean_std'] = Variable_pre_exp.ext_mass.apply(lambda x:x[1])
    Variable_pre_exp['noext_min_max'] = Variable_pre_exp.ext_mass.apply(lambda x:x[2])
    
    # 拼接
    Variable_pre_exp = pd.concat([Variable_pre_exp.set_index('var', drop=False), _desc, 
                                  pd.concat([_isnull_n,_nunique,_zero_n], axis=1)], axis=1, sort=False)
            
    # 取样
    if samples > 0:
        datain_sample = datain.apply(lambda x:pd.Series([np.nan] * samples, name=x.name) if \
                        np.sum(x.notna()) == 0 else x.dropna().sample(samples, replace = True).reset_index(drop = True))
        datain_sample = datain_sample.T.rename({i:'sample_' + str(i + 1) for i in range(samples)}, axis = 1)    
        
        Variable_pre_exp = pd.concat([Variable_pre_exp, datain_sample], axis=1, sort=False)    
        
    # 率延伸
    Variable_pre_exp['0_r'] = round(Variable_pre_exp['0_n'] * 100 / _N, 2).apply(str) + '%'
    Variable_pre_exp['na_r'] = round(Variable_pre_exp['na_n'] * 100 / _N, 2).apply(str) + '%'
    Variable_pre_exp['val_cate_r'] = round(Variable_pre_exp['val_cate_n'] * 100 / _N, 2).apply(str) + '%'
    Variable_pre_exp['num_r'] = round(Variable_pre_exp['count'] * 100 / (_N - Variable_pre_exp['na_n']), 2).apply(str) + '%'
    
    # 整理输出
    Variable_pre_exp = Variable_pre_exp[['var', 'var_type', 'na_n', 'na_r', 'val_cate_n', 
                                         'val_cate_r', 'mean_std', 'min_max', '0_n', '0_r', 
                                         'skew_kurt', _pvar, 'num_r','md_v', 
                                         'ext_low_upp_r', 'noext_mean_std', 'noext_min_max'
                                         ]
            + ['sample_' + str(i + 1) for i in range(samples)] if samples > 0 else []  # 取样字段 
            ]

    _rename = {'var': '字段名', 'var_type': '字段类型', 'min_max': '最小最大值', 'mean_std': '均值标准差', 
               '0_n': '0值数', '0_r': '0值率', 'val_cate_n': '唯一值数', 'val_cate_r': '唯一值率', 'skew_kurt': '偏度峰度', 
               'na_n': '缺失数', 'na_r': '缺失率', 'md_v': '众数值', _pvar: '分布情况'+str(qtl), 'num_r': '字段数值占比',
               'ext_low_upp_r': '上下界极端值比例('+str(ext_t)+'倍标准差)', 'noext_mean_std': '去极端值均值标准差', 'noext_min_max': '去极端值最小最大值'
               }
    _rename.update({'sample_' + str(i + 1):'抽样' + str(i + 1) for i in range(samples)} if samples > 0 else {})

    Variable_pre_exp_name = Variable_pre_exp.rename(index=str, columns=_rename)

    print('>>>>> Data explore list(N = ' + str(_N) + ') <<<<<\n', Variable_pre_exp_name)

    # 计算观测缺失率
    if calc_obs_na:
        obs_miss_rate = datain.apply(lambda row: np.sum(
            row.isnull()) / datain.shape[1], axis=1)
        _value_counts = obs_miss_rate.value_counts().sort_index()

        _labels = [round(x * 100, 2) for x in _value_counts.index]
        _quants = [round(x * 100 / datain.shape[0], 2) for x in _value_counts]

        _width = 0.4
        _ind = np.linspace(0.5, 9.5, len(_value_counts))

        fig = plt.figure(figsize=(12, 6))
        ax = fig.add_subplot(111)

        ax.bar(_ind - _width/2, _quants, _width)
        ax.set_xticks(_ind - _width/2)
        ax.set_xticklabels(_labels)

        ax.set_xlabel('Missing Rate(%)')
        ax.set_ylabel('Obs Rate(%)')
        ax.set_title('Observation missing rate')
        plt.grid()
        plt.show()
    else:
        obs_miss_rate = []

    # 1.2 特征分布、阈值¹
    if len(plt_out):
        Variable_pre_exp_name.to_csv(
            plt_out + '.csv', encoding=ecd, index=False)

        mpl.rcParams['axes.unicode_minus'] = False
        mpl.rcParams['font.sans-serif'] = ['SimHei']
        plt.rcParams['font.sans-serif'] = ['SimHei']

        try:
            with PdfPages(plt_out + '.pdf') as pdf:
                # 9宫格or4宫格
                sps = 4
                plt_control = {4: {'cols': 221, 'wspace': 0.4, 'hspace': 0.6, 'fontsize': 10},
                            9: {'cols': 331, 'wspace': 0.6, 'hspace': 0.8, 'fontsize': 8}}
    
                _cols = nume_cols | set(Variable_pre_exp.loc[(Variable_pre_exp.var_type == 'Character') & (
                    Variable_pre_exp.val_cate_n <= char_cate_threshold), 'var'])
    
                for t, x in enumerate(_cols):
                    plt.subplot(plt_control[sps]['cols'] + t % sps)
                    plt.subplots_adjust(
                        hspace=plt_control[sps]['hspace'], wspace=plt_control[sps]['wspace'])
    
                    if x in nume_cols:
                        datain[x].hist(bins=25)
                        plt.title('Num var ' + x + "'s distribution.",
                                fontsize=plt_control[sps]['fontsize'])
                    else:
                        char_vc = datain[x].value_counts()
                        plt.pie(char_vc.values, labels=char_vc.index,
                                autopct='%1.1f%%', startangle=90)
                        plt.axis('equal')
                        plt.title('Char var ' + x + "'s value counts.",
                                fontsize=plt_control[sps]['fontsize'])
    
                    if (t % sps == sps - 1) or (t == len(_cols) - 1):
                        pdf.savefig()
                        if plt_show:
                            plt.show()
                        plt.close()
        except:
            print('Notice:No plotting frameworks exist.')

    return Variable_pre_exp, obs_miss_rate



# -*- coding: utf-8 -*-
# Author: tigflanker
# Date: 21 Oct 2018
# Update: 18 Dec 2018
# 缺失值填补

# 导包/宏参设定
import pandas as pd 
import numpy as np

def Missing_Data_Impute(datain  # 需填补DF
                        ,imp_config = 'imp_mean'  # 填补方式，用法如上
                        ,define_key_parameter = ['imp_mean', 'imp_median', 'imp_mode', 'imp_knn']  # 手动设置关键值，如想填补的值和以上命令有重复
                        ):
    
    '''
    # 参数imp_config：
    # 0. 本宏采用的填补方法关键字如下：'imp_mean'/'imp_median'/'imp_mode'/'imp_knn'，'imp_knn'未启用
    # 1. 传入关键值，则数值型特征按照命令方式填补，字符型特征按照众数填补；如imp_config = 'imp_median'
    # 2. 如传入数值，则对输入数据集所有需要填补的连续性变量按照该数值填补，字符型变量按照众数填补；如imp_config = -999
    # 3. 如传入字典，则按照字典指定规则进行填补。
    #    最复杂情况例如：Age和Sex填均值、Income填中位数、Debt填-1，Class填'Unknow'
    #    则imp_config = {'imp_mean':['Age','Sex'], 'imp_median':'Income', -1:'Debt', 'Unknow':'Class'}
    
    # Example:
    datain_test = pd.DataFrame([[0,1,2,3,4,'a'],[5,6,7,8,9,'b'],[np.nan,np.nan,np.nan,np.nan,np.nan,np.nan]],
                        columns=['Index','Age','Sex','Income','Debt','Class'])
        
    imp_dataout = Missing_Data_Impute(datain_test)  
    imp_dataout = Missing_Data_Impute(datain_test, imp_config=-1)
    imp_dataout = Missing_Data_Impute(datain_test, imp_config={'imp_median':['Age','Sex'], 0:'Debt', 'Unknow':'Class'})
    '''
    
    # 字段类型列表
    datatp = datain.dtypes
    nume_cols = set(list(datatp[datatp != 'object'].index))
    char_cols = set(list(datatp[datatp == 'object'].index))
    
    # 参数解析
    # 解析出需填补字段
    need_to_imp = set(datain.columns[pd.isnull(datain).apply(sum) > 0])

    # 重写imp_config参数
    _imp_config = {}
    if imp_config in define_key_parameter:  # 默认全填补 
        if len(need_to_imp & nume_cols) > 0:
            _imp_config[imp_config] = need_to_imp & nume_cols
        if len(need_to_imp & char_cols) > 0:
            _imp_config[define_key_parameter[2]] = need_to_imp & char_cols
    elif type(imp_config) in (int, float):  # 指定值全填补
        if len(need_to_imp & nume_cols) > 0:
            _imp_config[imp_config] = need_to_imp & nume_cols
        if len(need_to_imp & char_cols) > 0:
            _imp_config[define_key_parameter[2]] = need_to_imp & char_cols
    if type(imp_config) != dict:
        imp_config = _imp_config
    print('本次填补规则如下：', imp_config)

    # 分析配置字典并开始填补   
    imp_dataout = datain.copy()
    
    # 填补
    for imp_value, imp_cols in imp_config.items():
        imp_cols = [imp_cols] if type(imp_cols) is str else imp_cols
        for imp_col in imp_cols:
           if imp_value == define_key_parameter[0]:
               imp_dataout[imp_col] = imp_dataout[imp_col].fillna(imp_dataout[imp_col].mean())  # 均值填补
           elif imp_value == define_key_parameter[1]:
               imp_dataout[imp_col] = imp_dataout[imp_col].fillna(imp_dataout[imp_col].median())  # 中位数填补
           elif imp_value == define_key_parameter[2]:
               imp_dataout[imp_col] = imp_dataout[imp_col].fillna(imp_dataout[imp_col].mode()[0])  # 众数填补
           else:
               imp_dataout[imp_col] = imp_dataout[imp_col].fillna(imp_value)  # 指定数填补
        
    return imp_dataout



# -*- coding: utf-8 -*-
# Author: tigflanker
# Date: 20 Oct 2018
# PDO转换: 由训练好的逻辑回归模型，按照PDO转化规则计算PDO评分 
# PDO_Score_trans宏用于计算每个特征分箱对应的PDO评分，需要输入为fa_df（包括分箱规则和对应WOE）、训练集（用途仅为计算ks最佳切点，用于计算好坏比）
# PDO_Score宏用于计算输入数据阵的每个特征的PDO评分和总评分，fa_df为必要参数（需提取分箱规则和PDO Score）

# 导包/宏参设定
import pandas as pd 
from pandas._libs import lib
import numpy as np
from scipy import stats 
import matplotlib.pyplot as plt
import math
from sklearn.metrics import roc_curve

# 4.3 PDO转换⁸ 
def PDO_Score(
        X_train  # 训练集
        ,y_train  # 训练集标签
        ,clf  # 模型
        ,fa_df
        ,pdo_p0 = 600  # pdo基线分 
        ,pdo_pdo = 20  # odds翻倍所需增加分值
        ,odds = 0  # odds部分讨论不完备，这里给出odds值用作直接定义
        ):

    # 1.设定总假设基准分 pdo_p0 pdo_pdo
    
    # 2.估算 AB
    # 2.1 按ks最佳切点计算总体odds值
    if not odds:  # 如果没有定义odds，将按照ks的百分位点区间内，去计算好坏比
        y_perdict = pd.Series(clf.predict_proba(X_train)[:,1])
        fpr, tpr, thresholds = roc_curve(y_train.reset_index(drop=True), y_perdict, pos_label = 1)
        w = tpr - fpr
        ks_cut = fpr[w.argmax()]  # ks_score = w.max()
        
        ks_true = y_train.reset_index(drop=True)[round(y_perdict, 2) == round(ks_cut, 2)]
        
        odds =  (1 - ks_true.mean()) / ks_true.mean()
    
    # 2.2 计算 AB值
    B = pdo_pdo / math.log(2)
    A = pdo_p0 - B * math.log(odds)
    
    # 2.3 计算每个变量、每个分箱的score 
    Basic_score = A + B * clf.intercept_[0]
    
    coef_dict = dict(zip(X_train.columns,clf.coef_[0]))
    
    fa_df['Score'] = list(map(lambda row:row[1] * B * coef_dict[row[0]], np.array(fa_df[['Var','Woe']])))

    return fa_df, Basic_score

# PDO打分
# 打分数据需提前填补 
def PDO_Score_Convert(
        PDO_datain
        ,fa_df
        ,y_var = ''
        ,Basic_score = 600
        ,num_bins = 50
        ):

    if set(fa_df.Var.unique()) != set(PDO_datain.columns) - set([y_var]):
        print('Columns not match, please check. This time use intersection set replace.')
    calc_cols = list(set(fa_df.Var.unique()) & set(PDO_datain.columns))
        
    def _sub_func(c):
        _var_score = np.array(fa_df[['Bins','Score']][fa_df.Var == c.name])   
        def _get_score(v):
            try:
                return list(filter(lambda x: v in x[0] if type(x[0]) is pd._libs.interval.Interval else v == x[0], _var_score))[0][1]
            except:
                return 0
            
        return pd.Series(lib.map_infer(c.astype(object).values, _get_score), name = c.name)
    
    PDO_Score_out = PDO_datain[calc_cols].apply(_sub_func)
    PDO_Score_out['PDO_Score_Convert'] = PDO_Score_out.apply(sum, axis = 1) + Basic_score
    
    # 频率分布图、T检验  
    if len(y_var) > 0:
        _pdo_y = pd.concat([PDO_datain[y_var], PDO_Score_out['PDO_Score_Convert']], axis=1, join='inner')
        
        x0 = _pdo_y.loc[_pdo_y[y_var] == 0,'PDO_Score_Convert']
        x1 = _pdo_y.loc[_pdo_y[y_var] == 1,'PDO_Score_Convert']
        
        P_value = stats.ttest_ind(x0,x1)[1]  
                
        fig, ax = plt.subplots()
        
        n, bins0, patches = ax.hist(x0, num_bins, density=1, histtype='step')
        n, bins1, patches = ax.hist(x1, num_bins, density=1, histtype='step')
        
        y0 = ((1 / (np.sqrt(2 * np.pi) * x0.std())) * np.exp(-0.5 * (1 / x0.std() * (bins0 - x0.mean()))**2))
        y1 = ((1 / (np.sqrt(2 * np.pi) * x1.std())) * np.exp(-0.5 * (1 / x1.std() * (bins1 - x1.mean()))**2))
        
        ax.plot(bins0, y0, '--')
        ax.plot(bins1, y1, '--')
        ax.set_xlabel('PDO Score')
        ax.set_ylabel('Probability density')
        ax.set_title('Histogram for 2 groups\nTtest P value:'+str(round(P_value,4)))
        
        fig.tight_layout()
        plt.show()   

    return PDO_Score_out

# 参考文章：https://blog.csdn.net/sscc_learning/article/details/78591210?utm_source=blogxgwz0

# 使用示例暂略，可参见Titanic测试实例


# -*- coding: utf-8 -*-
# Author: tigflanker
# Date: 31 Oct 2018
# Update: 04 Nov 2018：Add oversampling function 

import re
import pandas as pd 
import numpy as np

def Data_Sampling(datain,
                  sample_rule = {},
                  stratify_by = [],
                  group_by = ''
                  ): 
                  
    '''
    # 1. 单纯抽样：全部数据按照一定比例或指定数量进行抽样：sample_rule = 100; sample_rule = 0.3
    # 2. 分层抽样：按照某个变量进行排序，按照等频分箱，每箱内抽取均匀样本量：stratify_by = ['age', 20] + sample_rule = 100 or 0.3
    # 3. 非均衡抽样：① 设定某个分组变量，对其包含值个数取最小，按此最小值对所有分组 1:1 抽样：sample_rule = 'align' + group_by = 'survived'  
    #              ② 指定规则分组抽样，按照分组变量的每个值设定规则（如：100、0.3、'max'）：sample_rule = {0:100, 1:'max'} + group_by = 'survived'
    #              ③ 非均衡 + 分层抽样：sample_rule = {0:100, 1:0.3} + group_by = 'survived' + stratify_by = ['age', 20]
    # 系数中若倍数大于1，或例数大于样本量本身，即为过采样 
    
    # 使用示例：
    datain = pd.read_csv('http://biostat.mc.vanderbilt.edu/wiki/pub/Main/DataSets/titanic.txt')
    
    #print(datain.describe())
    #print(datain['survived'].value_counts())
        
    # 1. 简单抽样：sample_rule
    sample_out = Data_Sampling(datain,
                    sample_rule = 2.3,
                    stratify_by = [],
                    group_by = '')
    
    print(sample_out.describe())
    print(sample_out['survived'].value_counts())
    
    # 1. 简单抽样：sample_rule
    sample_out = Data_Sampling(datain,
                    sample_rule = 2000,
                    stratify_by = [],
                    group_by = '')
    
    print(sample_out.describe())
    print(sample_out['survived'].value_counts())
    
    # 2. 分层抽样：sample_rule + stratify_by
    sample_out = Data_Sampling(datain,
                    sample_rule = 2.3,
                    stratify_by = ['age', 20],
                    group_by = '')
    
    print(sample_out.describe())
    print(sample_out['survived'].value_counts())
    
    # 3. 非均衡抽样：sample_rule + group_by
    sample_out = Data_Sampling(datain,
                    sample_rule = 'align',
                    stratify_by = [],
                    group_by = 'survived')
    
    print(sample_out.describe())
    print(sample_out['survived'].value_counts())
    
    # 3. 非均衡抽样：sample_rule + group_by
    sample_out = Data_Sampling(datain,
                    sample_rule = {0:'max',1:4000},
                    stratify_by = [],
                    group_by = 'survived')
    
    print(sample_out.describe())
    print(sample_out['survived'].value_counts())
    
    # 3. 非均衡抽样：sample_rule + group_by
    sample_out = Data_Sampling(datain,
                    sample_rule = {0:2.0,1:4000},
                    stratify_by = ['age', 20],
                    group_by = 'survived')
    
    print(sample_out.describe())
    print(sample_out['survived'].value_counts())
    '''

    # Copy to final output dataset. 
    sample_out = datain.copy()
    os_times = 0 # Initialize rounds of oversampling.
    
    # Sub-macro for stratified sampling.
    def stratify_sample(sample_set_in = 'sample_out', 
                        stratify_n = 'stratify_by[1]', 
                        _stratify_by = 'stratify_by[0]', 
                        stratify_rule = 'sample_rule'):
        sample_set_in = sample_set_in.sort_values(_stratify_by).reset_index(drop = True)
        
        def _stra(l):
            _sample_rate = stratify_rule if type(stratify_rule) is float \
                                        else min(stratify_rule / stratify_n / l, 1)
            return list(np.random.choice([True,False],size=l,p=[_sample_rate,1-_sample_rate]))
        
        _bkt_n = [0] * stratify_n
        for x in range(len(sample_set_in)):
            _bkt_n[x%stratify_n] += 1
    
        keep_index = list(map(_stra, _bkt_n))
        keep_index = eval('[' + re.sub('[\[\]]','',str(keep_index)) + ']')  # list flattening
        while (type(stratify_rule) is not float) and (sum(keep_index) != stratify_rule):
            keep_index = list(map(_stra, _bkt_n))
            keep_index = eval('[' + re.sub('[\[\]]','',str(keep_index)) + ']')  # list flattening
        
        return keep_index
    
    # 均衡抽样 
    if type(sample_rule) not in (dict, str):
        
        # 提取过采样需求 
        os_times = int(sample_rule) if type(sample_rule) is float else sample_rule // len(sample_out)
        sample_rule = sample_rule % 1 if type(sample_rule) is float else sample_rule % len(sample_out)
        
        # 单纯抽样
        if (type(sample_rule) is float) and (len(stratify_by) == 0):
            print('>>> 当前正在执行简单抽样，抽样比例为：',os_times + sample_rule,'（过采样）' if os_times > 0 else '')
            keep_index = np.random.choice([True,False],
                                          size=len(sample_out),p=[sample_rule,1-sample_rule])
        if (type(sample_rule) is not float) and (len(stratify_by) == 0):  
            print('>>> 当前正在执行简单抽样，定样数量为：',os_times * len(sample_out) + sample_rule,'（过采样）' if os_times > 0 else '')
            keep_index = []
            while sum(keep_index) != sample_rule:
                keep_index = np.random.choice([True,False],
                                              size=len(sample_out),p=[sample_rule/len(sample_out),
                                                      1-sample_rule/len(sample_out)])
        
        # 分层抽样 
        if len(stratify_by) > 0:        
            print('>>> 当前正在执行分层抽样，分层变量为：',stratify_by[0],'，层数：',stratify_by[1],
                  '，抽样比例为：' if type(sample_rule) is float else '，定样数量为：',
                  os_times + sample_rule if type(sample_rule) is float else os_times * len(sample_out) + sample_rule,
                  '（过采样）' if os_times > 0 else '')
            keep_index = stratify_sample(sample_set_in = sample_out, stratify_n = stratify_by[1], _stratify_by = stratify_by[0], stratify_rule = sample_rule)
            
        # 均衡抽样部分的抽样实施动作 
        _sample_out = sample_out[keep_index]
        for _os in range(os_times):
            _sample_out = pd.concat([_sample_out, sample_out], axis=0, ignore_index=True, sort=False)
            
        sample_out = _sample_out
        
    # 非均衡抽样
    else :
        if sample_rule == 'align':
            _group_vc = sample_out[group_by].value_counts()
            _group_vc[:] = _group_vc.min()
            sample_rule = dict(_group_vc)
            print('>>> 已选择对齐方式，将按照',group_by,'组值中最小例数标齐各组例数，此例数为：',_group_vc.min())
            
        _sample_out = pd.DataFrame()
        for _value in sample_rule:
            # 划分数据集
            _sub_sample_set = sample_out.loc[sample_out[group_by] == _value]
            
            # 提取过采样需求
            if sample_rule[_value] != 'max':
                os_times = int(sample_rule[_value]) if type(sample_rule[_value]) is float else sample_rule[_value] // len(_sub_sample_set)
                sample_rule[_value] = sample_rule[_value] % 1 if type(sample_rule[_value]) is float else sample_rule[_value] % len(_sub_sample_set)
            
            # case
            if sample_rule[_value] == 'max':
                print('>>> 当前正在执行分组抽样，当前组值为：',group_by,'=',_value,'，按全量保留')
                keep_index = [True] * len(_sub_sample_set)
            elif len(stratify_by) == 0:
                if type(sample_rule[_value]) is float:
                    print('>>> 当前正在执行分组抽样，当前组值为：',group_by,'=',_value,'，抽样比例为：',os_times + sample_rule[_value],
                          '（过采样）' if os_times > 0 else '')
                    keep_index = np.random.choice([True,False],
                                                  size=len(_sub_sample_set),p=[sample_rule[_value],1-sample_rule[_value]])
                if type(sample_rule[_value]) is not float:
                    print('>>> 当前正在执行分组抽样，当前组值为：',group_by,'=',_value,'，定样数量为：',
                          os_times * len(_sub_sample_set) + sample_rule[_value],'（过采样）' if (os_times > 0) & (sample_rule[_value] > 0) else '')
                    keep_index = []
                    while sum(keep_index) != sample_rule[_value]:
                        keep_index = np.random.choice([True,False],
                                                      size=len(_sub_sample_set),p=[sample_rule[_value]/len(_sub_sample_set),
                                                              1-sample_rule[_value]/len(_sub_sample_set)])
            elif len(stratify_by) > 0:
                print('>>> 当前正在执行分组 + 分层抽样，当前组值为：',group_by,'=',_value,
                      '，分层变量为：',stratify_by[0],'，层数：',stratify_by[1],
                      '，抽样比例为：' if type(sample_rule[_value]) is float else '，定样数量为：',
                      sample_rule[_value] + os_times if type(sample_rule[_value]) is float else os_times * len(_sub_sample_set) + sample_rule[_value],
                      '（过采样）' if os_times > 0 else '')
                keep_index = stratify_sample(sample_set_in = _sub_sample_set, stratify_n = stratify_by[1], 
                                             _stratify_by = stratify_by[0], stratify_rule = sample_rule[_value])
            
            __sub_sample_set = _sub_sample_set[keep_index]
            for _os in range(os_times):
                __sub_sample_set = pd.concat([__sub_sample_set, _sub_sample_set], axis=0, ignore_index=True, sort=False)
            _sub_sample_set = __sub_sample_set
            
            _sample_out = pd.concat([_sample_out, _sub_sample_set], axis=0, ignore_index=True, sort=False)
    
        sample_out = _sample_out
        
    return sample_out
    


# -*- coding: utf-8 -*-
# Author: tigflanker
# Date: 10 Dec 2018
# Desc: 本方法用于观察数据某一列变量的分布情况，用于策略定制时的节点划分

import pandas as pd 
import numpy as np
import matplotlib.pyplot as plt

def Data_Distribution(
        serin = ''  # 输入序列
        ,bins = []  # 手动分箱list
        ,closed = 'left'
        ):
    
    serin = serin.dropna()

    bins.insert(0, serin.min())
    bins.append(serin.max())
    
    fig, (ax, ax0, ax1) = plt.subplots(ncols=3, figsize=(8, 8))
    
    ax0 = plt.subplot(221) # 第一行的左图
    ax1 = plt.subplot(222) # 第一行的右图
    ax  = plt.subplot(212) # 第二整行
    
    _value_counts = serin.value_counts().sort_index()
    ax.plot(_value_counts.index, list(map(lambda x:sum(_value_counts[_value_counts.index <= x])/sum(_value_counts),_value_counts.index)))
 
    ax.grid()
    ax.set_title('Cumu')
    ax.set_xticks(np.linspace(serin.min(),serin.max(),15))
    ax.set_xlabel('Value')
    ax.set_ylabel('Cumulative percentage')
 
    ax0.hist(serin, bins = bins, histtype='bar', rwidth=0.8, cumulative=False)
    ax0.set_xticks(bins)
    ax0.set_title('Hist') 
    
    # 计算分布矩阵
    _bins = list(map(lambda i:tuple([bins[i],bins[i+1]]),range(len(bins) - 1)))
     
    # table
    _bins = pd.IntervalIndex.from_tuples(_bins, closed = closed)
    _vc = pd.cut(serin, _bins).value_counts()

    _table = pd.DataFrame(list(zip(list(_vc.index),list(_vc))),columns=['interval','N'])
    _table = _table.sort_values('interval').reset_index(drop = True)
    
    # Calc
    _R, _cum_N, _cum_R = 0, 0, 0
    _calc_list = []
    for x in _table['N']:
        _cum_N += x
        _R = x * 100 / sum(_table['N'])
        _cum_R += _R
        
        _calc_list.append(['%.2f%%' % _R, _cum_N, '%.2f%%' % _cum_R])
        
    _calc_list = pd.DataFrame(_calc_list, columns=['rate','cum N','cum Rate'])
    _table = pd.merge(_table, _calc_list, left_index=True, right_index=True)  
    # 分布矩阵计算完毕

    ax1.pie(list(_table['N']),labels=list(_table['interval']),
            autopct='%1.1f%%',shadow=False,startangle=90) 
    ax1.axis('equal')  
    ax1.set_title('Pie')

    fig.tight_layout()
    plt.show()

    print(_table)
    
if __name__ == '__main__':
    datain_example = pd.read_csv('http://biostat.mc.vanderbilt.edu/wiki/pub/Main/DataSets/titanic.txt')
    
    Data_Distribution(serin = datain_example['age'], bins = [15, 60])  # 年龄



# -*- coding: utf-8 -*-
# Author: tigflanker
# Date: 29 Dec 2018

from statsmodels.stats.outliers_influence import variance_inflation_factor
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns

def Corr_Vif(
        datain
        ,corr_method = 'pearson'  # 'pearson', 'kendall', 'spearman'
        ,corr_plt_filter = 20
        ,font_size = 12
        ,figsize = (14, 10)
        ,rounds = 2
        ):
        
    '''
    # Example:
    from sklearn.datasets import load_breast_cancer
        
    datain_example = load_breast_cancer()
    datain_example = pd.DataFrame(datain_example.data, columns=datain_example.feature_names)
    
    corr_df = Corr_Vif(datain_example, corr_plt_filter=0.95, figsize=(18,12))
    corr_df = Corr_Vif(datain_example, corr_plt_filter=15, figsize=(18,12))    
    '''

    # VIF(方差膨胀因子)
    _values = datain.values
    vif_list = list(map(lambda x: [datain.columns[x], variance_inflation_factor(
        _values, x)], range(len(datain.columns))))
    vif_list = pd.DataFrame(vif_list, columns=['Var', 'Vif'])

    # Corss table
    corr_table = datain.corr(method = corr_method).fillna(0)
    
    # 计算需要展示的变量 
    dict_x = {}
    for c, row in corr_table.iteritems():
        for v in row.index:
            if c != v and row[v] > (corr_plt_filter if type(corr_plt_filter) is float else 0):
                dict_x[(c, v)] = abs(row[v])
            
    show_cols = set()
    for x in sorted(dict_x.items(),key = lambda x:x[1],reverse = True):
        show_cols.add(x[0][0])
        show_cols.add(x[0][1])
        
        if len(show_cols) >= (corr_plt_filter if type(corr_plt_filter) is int else 100):
            break    
    
    if len(show_cols) >= 2:
        f, ax = plt.subplots(figsize = figsize)
        sns.heatmap(corr_table[list(show_cols)].T[list(show_cols)].applymap(lambda x: round(x, rounds)), annot=True,
                    annot_kws={'size': font_size}, linewidths=0.05)
        ax.tick_params(axis='x', labelsize=font_size)
        ax.tick_params(axis='y', labelsize=font_size)
        plt.title('Character variable CORR heat map('+corr_method+').')
        plt.show()

    return pd.merge(vif_list, corr_table, left_on='Var', right_index=True)



# -*- coding: utf-8 -*-
# Author: guoyang14
# Date: 27 May 2019
# 敏感值加密解密

import random 
import hashlib
import pandas as pd

# 函数用途为对数据集中敏感信息加密，例如对模型标签加密
# PSR参数为加密/解密字典，key必须包含：struct、r、s、p（p为解密过程必须参数）
#    struct：结构排列方式
#    r：random，为随机数位数，默认为千位内
#    s：salt，为密令
#    p：加密过程留空，解密过程指定，例如 p:[0, 1]
#    例：解密过程：PSR = {'struct':'psr','r':3,'s':'apple','p':['AS','DF','AD']}
def PSR_Crypto(
        Serin = ''  # 输入类型为序列型
        ,PSR = {'struct':'psr','r':3,'s':'apple'}  # P:password,R:random,S:salt；P定义则为解密过程
        ,Serout = ''  # 加密/解密后字段命名，留空则设置为原字段名加固定后缀
        ):
        
    '''
    # Example
    exam_ser = pd.Series(map(lambda x:random.choice(['X0', 'X1']), range(100000)), name = 'col')
    
    # 加密后
    %time exam_encrypt = PSR_Crypto(Serin = exam_ser, PSR = {'struct':'psr','r':3,'s':'apple'})
    
    # 解密后
    %time exam_origin = PSR_Crypto(Serin = exam_encrypt, PSR = {'struct':'psr','r':3,'s':'apple','p':['X0', 'X1']})
    
    # 检查加密前后是否一致
    exam_check = pd.concat([exam_ser, exam_encrypt, exam_origin], axis = 1)
    sum(exam_check.col != exam_check.col_encrypt_origin)
    '''

    def hm(s):
        hm = hashlib.md5()
        hm.update(s.encode('utf8'))
        
        return hm.hexdigest()

    if 'p' in PSR:
        # 解密过程
        _pl = PSR['p']
        
        def _psr(_p):
            _s = PSR['s']
            _e = ' + '.join(['str(_' + i + ')' for i in PSR['struct']])
            
            return pd.DataFrame(list(map(lambda _r:[hm(eval(_e, {'_s': _s, '_r': _r, '_p': _p})), _p], range(10 ** PSR['r'])))
                                , columns = [Serin.name, Serout if len(Serout) else Serin.name + '_origin'])
            
        return pd.merge(pd.DataFrame(Serin), pd.concat(map(_psr, _pl)), left_on = Serin.name, right_on = Serin.name, how = 'left')
     
    else:
        # 加密过程        
        def _psr(t):
            _p = t
            _s = PSR['s']
            _r = random.randint(0, 10 ** PSR['r'] - 1)
        
            return hm(eval(' + '.join(['str(_' + i + ')' for i in PSR['struct']])))
    
        print('Encrypt process, struct:'+PSR['struct']+', random digit:'+str(PSR['r'])+', salt:'+str(PSR['s']))
        return pd.Series(Serin.map(_psr), name = Serout if len(Serout) else Serin.name + '_encrypt')
    



# -*- coding:utf-8 -*- 

from rcmf.Data_Explore import Data_Explore
from rcmf.Missing_Data_Impute import Missing_Data_Impute
from rcmf.Dummy import Dummy
from rcmf.Chi2_merge import calc_chi2, Chi2_Merge
from rcmf.Cut_Merge import Cut_Merge
from rcmf.Woe_Iv import Woe_Iv
from rcmf.Corr_Vif import Corr_Vif
from rcmf.PSI import psi, PSI
from rcmf.PDO_Score import PDO_Score, PDO_Score_Convert
from rcmf.Model_Performance import Model_Performance
from rcmf.Data_Sampling import Data_Sampling
from rcmf.Data_Distribution import Data_Distribution
from rcmf.PSR_Crypto import PSR_Crypto

__version__ = '1.0.6'

__all__ = (
    Data_Explore,
	Missing_Data_Impute, Dummy, calc_chi2, Chi2_Merge, Cut_Merge, Woe_Iv, Corr_Vif, psi, PSI, 
	PDO_Score, PDO_Score_Convert,
	Model_Performance,
	Data_Sampling, Data_Distribution,
    PSR_Crypto
	)
	


# -*- coding: utf-8 -*-
# Author: tigflanker
# Date: 04 Oct 2018
# Update: 10 Dec 2018：修正分箱边界问题 
# 卡方分箱

import re
import pandas as pd 
import numpy as np
from scipy import stats
import math
import matplotlib.pyplot as plt

def calc_chi2(datain, group = 'group', y_var = 'y_var'):
    N = len(datain)
    chi2 = 0
    for g in set(datain[group]):
        for y in set(datain[y_var]):
            Agj = sum((datain[group] == g) & (datain[y_var] == y))
            Egj = sum(datain[group] == g) * sum(datain[y_var] == y) / N
            chi2 += 0 if Egj == 0 else (Agj - Egj)**2/Egj

    return chi2

def Chi2_Merge(
        chi2_datain = 'datain'
        ,divided_var = 'divided_var'
        ,y_var = 'y_var'
        ,aim_groups = 'auto'  # 目标组数(手动指定卡方分箱的目标组数，不建议改动)
        ,step_size = 1  # 该参数用于定义每次迭代收集最小的N个卡方值对应的组标进行合并，用于提高合并速率，不建议大于5
        ,print_detail = False
        ,calc_entropy = False
        ,collection_point = 20  # 在分组降维至该参数以下时，收集卡方值/收集分桶规则
        ):

    chi2_datain = chi2_datain.loc[chi2_datain[divided_var].notna()].sort_values(by=[divided_var, y_var]).reset_index()
    chi2_datain[divided_var + '_chi2_group'] = chi2_datain.index
    Total_variable_length = len(chi2_datain)
    
    chi2_change = []  # 用于绘制卡方和值变化图
    divide_rule = []  # 用于收集区间规则
    spr_list = [] # 收集Spearman-R变化
    while(len(chi2_datain[divided_var + '_chi2_group'].unique()) > 2):
        group_list = list(chi2_datain[divided_var + '_chi2_group'].unique())
        
        min_chi2 = {}  # 用于收集卡方值及其组标
        chi2_avg = 0
        if len(chi2_datain[divided_var + '_chi2_group'].unique()) <= 30:  # 30以内降速为1，以防跨过目标组数
            step_size = 1
        for i in range(len(group_list) - 1):
            chi2 = calc_chi2(chi2_datain[(chi2_datain[divided_var + '_chi2_group'] == group_list[i]) \
                                         | (chi2_datain[divided_var + '_chi2_group'] == group_list[i + 1])]
                             , group = divided_var + '_chi2_group', y_var = y_var)  
            chi2 = round(chi2, 8)  # 将计算得到的卡方值按8位精度截尾
            chi2_avg += chi2
            if chi2 in min_chi2:
                min_chi2[chi2] = min_chi2[chi2] + [group_list[i]]  
            elif len(min_chi2) < step_size:
                min_chi2[chi2] = [group_list[i]]  
            elif chi2 < max(min_chi2):
                del min_chi2[max(min_chi2)]
                min_chi2[chi2] = [group_list[i]]
                
        if len(min_chi2) < step_size:
            min_chi2 = {min(min_chi2) : min_chi2[min(min_chi2)]}
        merge_group_list = set(eval('[' + re.sub('[\[\]]','',str(list(min_chi2.values()))) + ']'))  # 数组扁平化
                        
        # 变组
        def temp_func1(x):
            if x in merge_group_list:
                x = group_list[group_list.index(x) + 1]
                while(x in merge_group_list):
                    x = group_list[group_list.index(x) + 1]
            return x 
        
        chi2_datain.loc[:,divided_var + '_chi2_group'] = list(map(temp_func1,chi2_datain[divided_var + '_chi2_group']))
        
        # 计算Spearman相关系数
        calc_spm = chi2_datain.groupby(divided_var + '_chi2_group')
        spm_r, spm_p = stats.spearmanr(calc_spm[divided_var + '_chi2_group'].mean(), calc_spm[y_var].mean())
        
        # 熵和gini(gini暂略)⁴
        if calc_entropy:
            def chi2_entropy(y):
                data_part = chi2_datain.loc[chi2_datain[divided_var + '_chi2_group'] == y,[y_var,divided_var + '_chi2_group']]
            
                return (len(data_part) / Total_variable_length) * \
                       sum(map(lambda x: - (sum(data_part[y_var] == x) / len(data_part)) * \
                               math.log((sum(data_part[y_var] == x) / len(data_part)),2)
                               ,data_part[y_var].unique()))
        
            chi2_entropy = sum(map(chi2_entropy, chi2_datain[divided_var + '_chi2_group'].unique()))
        
        # 输出
        if print_detail:
            if calc_entropy:
                chi2_entropy = ', entropy ' + str(round(chi2_entropy, 8))
            else:
                chi2_entropy = ''
                
            print('Tot-len',Total_variable_length
                  ,', cur-gps',len(chi2_datain[divided_var + '_chi2_group'].unique())
                  ,', step size',step_size
                  ,', avg chi2',round(chi2_avg / len(group_list), 4)
                  ,', spr-r',round(spm_r, 4)
                  ,chi2_entropy)
            
            # 收集卡方值的变化  
            if len(chi2_datain[divided_var + '_chi2_group'].unique()) <= collection_point:
                chi2_change.append([len(chi2_datain[divided_var + '_chi2_group'].unique())
                                    ,round(chi2_avg / len(group_list), 8)
                                    ,round(chi2_avg / len(group_list), 8) - last_chi2
                                    ,round(spm_r, 8)])
                
        # 收集Spearman-R变化 
        if len(chi2_datain[divided_var + '_chi2_group'].unique()) <= collection_point:
            spr_list.append(spm_r)                
        last_chi2 = round(chi2_avg / len(group_list), 8)  # 记录上次(变组前)卡方均值    
                
        # 重新序列划分好的区间值
        if len(chi2_datain[divided_var + '_chi2_group'].unique()) <= collection_point:
            current_group_id = list(chi2_datain[divided_var + '_chi2_group'].unique())
            chi2_datain[divided_var + '_chi2_group'] = list(map(lambda x:current_group_id.index(x) + 1, chi2_datain[divided_var + '_chi2_group']))
    
            # 划分规则
            divide_rule_x = chi2_datain.groupby(divided_var + '_chi2_group')[divided_var].agg(['min', 'max'])
            divide_rule_x['groups'] = len(divide_rule_x)
            
            if len(divide_rule):  
                divide_rule = pd.concat([divide_rule, divide_rule_x])                    
            else:
                divide_rule = divide_rule_x
    
    # 绘制卡方变化图⁵ 
    if print_detail:
        x_group = list(map(lambda x:x[0], chi2_change))
        fig = plt.figure()
        
        ax1 = fig.add_subplot(111)
        ax1.plot(x_group, list(map(lambda x:x[3], chi2_change)))
        ax1.set_xlabel('Groups change')
        ax1.set_ylabel('Spearman R change(blue)')
        ax1.set_title('Spearman R change')
        
        ax2 = ax1.twinx()  # this is the important function
        ax2.plot(x_group, list(map(lambda x:x[1], chi2_change)), 'r')
        ax2.set_ylabel('Chi2 value change(red)')
        
        plt.grid(axis='x')  # 失效了
        plt.xticks(x_group)
        plt.show()        
        
    # 给出最优分箱组数 
    # 规则：首先筛选有效分箱组；其次在有效分箱组中从多至少遍历，
    #      直至Spearman最佳分箱点，或者是最小有效分箱组
    divide_rule['interval'] = divide_rule['max'] - divide_rule['min']
    available_groups =  list(filter(lambda x:
        len(set(divide_rule.loc[divide_rule.groups == x,'interval']) - set([0])) > 1
        ,set(divide_rule.groups)))
        
    # Spearman-R指示最佳分箱点 
    spr_index_bst = len(spr_list) - np.argwhere(abs(np.array(spr_list)) == max(abs(np.array(spr_list)[:-1])))[0,0] + 1  
    
    best_cut_point = max(min(available_groups), spr_index_bst) if aim_groups == 'auto' else aim_groups
    
    _best_cut_rule = divide_rule.loc[divide_rule.groups == best_cut_point,['min','max']]
    best_cut_rule = []
    
    _best_cut_rule = _best_cut_rule[_best_cut_rule['max'] != _best_cut_rule['min']].reset_index(drop = True)
    for i in _best_cut_rule.index:
        if i == 0:
            best_cut_rule.append(tuple([-np.inf, (_best_cut_rule.loc[i,'max'] + _best_cut_rule.loc[i + 1,'min']) / 2 ]))
        elif i == max(_best_cut_rule.index):
            best_cut_rule.append(tuple([(_best_cut_rule.loc[i,'min'] + _best_cut_rule.loc[i - 1,'max']) / 2, np.inf]))
        else: #if _best_cut_rule.loc[i,'min'] != _best_cut_rule.loc[i,'max']:
            best_cut_rule.append(tuple([(_best_cut_rule.loc[i,'min'] + _best_cut_rule.loc[i - 1,'max']) / 2, 
                                       (_best_cut_rule.loc[i,'max'] + _best_cut_rule.loc[i + 1,'min']) / 2]))
            
    if print_detail:
        print(divided_var + "'s " + 'best chi2-merge groups:', len(best_cut_rule))
        
    # 按最优分箱点分箱 
    def cut_temp(x):
        group_x = 0
        for i, d in enumerate(best_cut_rule):
            if (-np.inf == d[0]) & (d[1] == x):
                group_x = i
                break
            elif (d[0] <= x) & (x < d[1]):  # 左闭右开，边界值往后堆
                group_x = i
                break
        return group_x
        
    chi2_datain[divided_var] = list(map(cut_temp, chi2_datain[divided_var])) 
    chi2_datain = chi2_datain[['index', divided_var]]
    chi2_datain.set_index(['index'], inplace=True)       

    return chi2_datain, best_cut_rule # divide_rule

if __name__ == '__main__':
    datain = pd.read_csv('http://biostat.mc.vanderbilt.edu/wiki/pub/Main/DataSets/titanic.txt')

    chi2_df, chi2_divide_rule = Chi2_Merge(chi2_datain = datain[['survived','age']], y_var = 'survived'
                                        , divided_var = 'age'
                                        , print_detail = True, step_size = 3)
    
    # ref 
    # ref1: https://www.powershow.com/view1/1fa37b-ZDc1Z/ChiMerge_Discretization_powerpoint_ppt_presentation
    # ref2: http://www.aaai.org/Papers/AAAI/1992/AAAI92-019.pdf
    # ref3: https://blog.csdn.net/autoliuweijie/article/details/51594373
    


# -*- coding: utf-8 -*-
# Author: tigflanker
# Date: 04 Oct 2018
# Update: 01 Jan 2019: fa_df在本程序中列为必须，WoE-IV可直接依照fa_df统计内容算出 
#                      如无使用Cut_merge分箱，请采用之前版本

# 0. 导包/宏参设定
import pandas as pd 
import math

# 2.3.1 WOE-IV
def Woe_Iv(datain, fa_df):

    '''
    # Example 
    
    # 数据读入
    # 1. 使用乳腺癌预测数据集 
    from sklearn.datasets import load_breast_cancer
    datain_example = load_breast_cancer()
    data_part = pd.DataFrame(datain_example.data, columns=datain_example.feature_names)
    flag_part = pd.Series(datain_example.target, name='y_var')
    datain_example = pd.concat([flag_part, data_part], axis=1)
    
    # 2. 使用泰坦尼克号数据 
    datain_example = pd.read_csv('http://biostat.mc.vanderbilt.edu/wiki/pub/Main/DataSets/titanic.txt')
    
    # 分箱计算
    cut_df, fa_df = Cut_Merge(datain_example, cut_way = 'ef', cut_n = 5, cut_y = 'y_var')
    
    # 测试
    wi_dataout, woe_dict, fa_df = Woe_Iv(cut_df, y_var = 'y_var', fa_df = fa_df)
    '''
    
    # 计算好坏总数
    _sum_set = fa_df.groupby('Var').agg({'bads':sum, 'N':sum}).rename(columns={'bads':'bads_sum','N':'N_sum'})
    fa_df = pd.merge(fa_df, _sum_set, how='left', left_on='Var', right_index=True)
    
    # WoE_IV计算主方法 
    def _WoE_IV(x):
        LS = int(x.bads in (0, x.N))
        
        PYi = (x.bads + LS) / (x.bads_sum + LS)
        PNi = (x.N - x.bads + LS) / (x.N_sum - x.bads_sum + LS)
            
        WoE = math.log(PYi / PNi)   
        return WoE, (PYi - PNi) * WoE
       
    fa_df = pd.concat([fa_df, pd.DataFrame(list(fa_df.apply(_WoE_IV, axis = 1)), columns=['Woe', 'Iv'])], axis=1) 
    _iv_set = pd.DataFrame(fa_df.groupby('Var').Iv.sum(), columns = ['Iv'])
    _iv_set = pd.concat([_iv_set, _iv_set.Iv.rank(ascending=False).rename('Iv_rank')], axis=1)
    fa_df = pd.merge(fa_df.drop(['Iv', 'bads_sum', 'N_sum'], axis=1), _iv_set, how='left', left_on='Var', right_index=True)    
    
    # 2.3.1.3 WOE数据集映射 
    wi_dataout = cut_df.copy()
    
    woe_dict = {}
    for col in set(fa_df.Var):
        woe_dict[col] = dict(zip(fa_df.Binn[fa_df.Var == col], fa_df.Woe[fa_df.Var == col]))

    wi_dataout.update(wi_dataout[list(set(fa_df.Var))].apply(lambda x:x.map(lambda y:woe_dict[x.name][y])))
    
    # 2.3.1.5 收尾（结果update，输出）
    return wi_dataout, woe_dict, fa_df



# -*- coding: utf-8 -*-
# Author: Huyao
# Packing: GY
# Inc Date: 12 Dec 2018

import math
import numpy as np
import pandas as pd 

# 单独PSI计算程序
def psi(
        base = pd.Series()
        ,comp = pd.Series()
        ,threshold = [10, 100]
        ,check = False
        ,drop_na = False
        ):
    
    # 基本统计/参数分解
    groups = threshold[0]
    max_groups = threshold[1]
        
    union_values = set(base.dropna()) | set(comp.dropna()) | set({} if drop_na else {np.nan})
    union_groups = len(union_values)
    base_n, comp_n = len(base), len(comp)
    
    # 箱内计算
    def bin_psi_calc(x):
        
        if type(x) in (tuple, list):
            base_count = np.sum(base.between(x[0], x[1]))
            comp_count = np.sum(comp.between(x[0], x[1]))
        else:
            base_count = np.sum(base == x)
            comp_count = np.sum(comp == x)
        
        lpls = int(base_count * comp_count == 0)
        
        base_ratio = (base_count + lpls) / (base_n + lpls)
        comp_ratio = (comp_count + lpls) / (comp_n + lpls)
    
        if check:
            print('Var:',base.name,'Bin:',x, 'Bin PSI:',round((base_ratio - comp_ratio) * math.log(base_ratio / comp_ratio), 2))
    
        return (base_ratio - comp_ratio) * math.log(base_ratio / comp_ratio)
    
    if 'object' in (base.dtype, comp.dtype):
        if union_groups > max_groups:
            print('The character type variable "'+base.name+'" contain too many value types(',union_groups,').')
            return np.nan
        else:
            return np.sum(list(map(bin_psi_calc, union_values)))
    else:  # 纯数值型
        if union_groups <= groups:
            return np.sum(list(map(bin_psi_calc, union_values)))  # 直接按照字符型对待
        else:        
            cut_points = np.linspace(min(union_values - {np.nan}), max(union_values - {np.nan}), num=groups + 1)
                    
            return np.sum(list(map(lambda u:bin_psi_calc((cut_points[u], cut_points[u + 1])), range(groups))) 
                          + ([] if drop_na else [bin_psi_calc(np.nan)]))
        
# 数据集循环计算(调度)
def PSI(base
        ,comp
        ,threshold = [10, 100]
        ,check = False
        ,drop_na = False
        ):
        
    """
    PSI计算说明：
    1. 如果比较对象为字符型/分类型变量，则需要按照两边的类数并集进行切割（是否需要考虑设置上限？）
    2. 如果比较对象为连续型变量，并且两边值类型数大于待划分类数，则正常划分并计算PSI
    3. 如果比较对象为连续型变量，两边并集类数也小于待划分类数，则取每个类数单独成组进行计算
    
    *1. 两边值的缺失值单独计算一组
    *2. 如若出现空箱，采用拉普拉斯平滑处理：当前两边组的例数均加1（分母同时加1）
    
    参数说明：
    统一使用下方的PSI函数，psi为子函数
    Base/Comp：两侧待比对数据，输入类型为Dataframe或Series
    threshold：为元组或列表，第一位为分箱个数，第二位针对字符型变量，如果超过类数，则不计算该变量
    check：检查，若关闭则静默计算
    
    # Example:
    from sklearn.datasets import load_breast_cancer
    datain_example = load_breast_cancer()
    datain_example = pd.DataFrame(datain_example.data, columns=datain_example.feature_names)
    
    psi_value = PSI(datain_example['mean radius'], datain_example['mean texture'])
    psi_value = PSI(datain_example[datain_example.index < 300], datain_example[datain_example.index >= 300])
    """

    if type(base) is pd.DataFrame and type(comp) is pd.DataFrame:
        com_cols = set(base.columns) & set(comp.columns)
        psi_value = list(map(lambda col:[col, psi(base[col], comp[col], threshold, check, drop_na)], com_cols))
        psi_value = pd.DataFrame(psi_value, columns = ['Var', 'Psi'])
    else:
        psi_value = psi(base, comp,  threshold, check, drop_na)
        print('PSI value:', psi_value)
        
    return psi_value



"""A setuptools based setup module.
See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject

Refer to:
https://github.com/pypa/sampleproject/blob/master/setup.py
https://github.com/ShichenXie/scorecardpy/blob/master/setup.py
https://www.cnblogs.com/yixianclove/p/7776792.html
"""

# Always prefer setuptools over distutils
from setuptools import setup #, find_packages
from os import path
# io.open is needed for projects that support Python 2.7
# It ensures open() defaults to text mode with universal newlines,
# and accepts an argument to specify the text encoding
# Python 3 only projects can skip this import
from io import open

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# Arguments marked as "Required" below must be included for upload to PyPI.
# Fields marked as "Optional" may be commented out.

setup(
    name='rcmf',  # Required
    version='1.0.6',  # Required

    description='A risk control modeling toolbox, include scattered function for building scorecard/machine learn model.',  # Optional
    long_description=long_description,  # Optional
    long_description_content_type='text/markdown',  # Optional (see note above)

    # url='https://github.com/tigflanker/Python/tree/master/Risk control modeling framework',  # Optional 
    author='Tigflanker & deng@jd partners',  # Optional
    author_email='tigflanker@qq.com',  # Optional

    classifiers=[  # Optional
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        # 'Topic :: Other :: Other',

        # Pick your license as you wish
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        # These classifiers are *not* checked by 'pip install'. See instead
        # 'python_requires' below.
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],

    keywords='scorecard, python, toolbox',  # Optional

    # You can just specify package directories manually here if your project is
    # simple. Or you can use find_packages().
    #
    # Alternatively, if you just want to distribute a single Python file, use
    # the `py_modules` argument instead as follows, which will expect a file
    # called `my_module.py` to exist:
    #
    #   py_modules=["my_module"],
    #
    packages=['rcmf'],  # Required

    # Specify which Python versions you support. In contrast to the
    # 'Programming Language' classifiers above, 'pip install' will check this
    # and refuse to install the project if the version does not match. If you
    # do not support Python 2, you can simplify this to '>=3.5' or similar, see
    # https://packaging.python.org/guides/distributing-packages-using-setuptools/#python-requires
    # python_requires=['numpy','pandas','scipy','matplotlib','seaborn','sklearn','scikit-plot','statsmodels'],

    # This field lists other packages that your project depends on to run.
    # Any package you put here will be installed by pip when your project is
    # installed, so they must be valid existing projects.
    #
    # For an analysis of "install_requires" vs pip's requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=['numpy','pandas','scipy','matplotlib','seaborn','sklearn','scikit-plot','statsmodels'],  # Optional

    # List additional URLs that are relevant to your project as a dict.
    #
    # This field corresponds to the "Project-URL" metadata fields:
    # https://packaging.python.org/specifications/core-metadata/#project-url-multiple-use
    #
    # Examples listed include a pattern for specifying where the package tracks
    # issues, where the source is hosted, where to say thanks to the package
    # maintainers, and where to support the project financially. The key is
    # what's used to render the link text on PyPI.
    # project_urls={  # Optional
        # 'Source': 'https://github.com/tigflanker/Python/tree/master/Risk control modeling framework/',
        # 'Say Thanks!': 'Thanks to my friends HuYao@jd & JiangChao@jd for their support on this project many times.',
    # },
) 


# -*- coding: utf-8 -*-
# Author: tigflanker
# Date: 19 Dec 2018
# Version: 1.0
# 说明：该宏为机器学习线，模型表现；输入对象为分隔好的训练和验证集，外加预测模型：
# X_train, X_test, y_train, y_test和clf 
# scikitplot为@Huyao推荐，方便绘制sklearn模型产出 

# 导包/宏参设定
import numpy as np
import pandas as pd 
import matplotlib.pyplot as plt
import scikitplot as skplt #  pip install scikit-plot
from sklearn.metrics import auc,roc_curve
from sklearn.metrics import confusion_matrix

# 4. 机器学习模型表现评定
def Model_Performance(clf  # 输入模型
                      ,X_train  # 训练集特征
                      ,y_train  # 训练集标签
                      ,X_test = []  # 测试集特征
                      ,y_test = []  # 测试集标签
                      ,out_path = ''  # 特征重要性输出，xlsx格式 
                      ,calc_list = ['pr','roc','ks']  # 需计算的项目列表，分别为：准召列表、ROC/AUC、KS、特征重要性、模型学习率['cut','pr','roc','ks','fi','lc']
                      ,calibration = {}  # calibration图，如{'Random Forest':clf_rf,'Logistic Regression':clf_lr}
                      ):
                      
    '''
    # Example:
    Model_Performance(clf_xgb, X_train, y_train, X_test, y_test, out_path = 'D:/xxx.xlsx', calibration = {'Random Forest':clf_rf,'XGboost':clf_xgb})
    Model_Performance(clf_xgb, X_train, y_train, calc_list = ['pr', 'roc', 'ks', 'cut'])
    '''
    
    # 4.1 切点排序及准召、F1-Score
    def PRTF(X = 'X_train', y = 'y_train', clf = 'clf'):
        sum_y = sum(y)
        sum_op_y = sum(y != 1)

        def _PRTF(i):
            conf_mat = confusion_matrix(y, clf.predict_proba(X)[:,1] > i)

            precision = round(conf_mat[1][1] / (conf_mat[0][1] + conf_mat[1][1]), 4) if conf_mat[0][1] + conf_mat[1][1] else 0
            recall = round(conf_mat[1][1] / sum_y, 4)  # recall = TPR 
            FPR = round(conf_mat[0][1] / sum_op_y, 4)

            return [i, precision, recall, FPR, round(2 * precision * recall / (precision + recall),4)]

        pred_max = max(clf.predict_proba(X)[:,1])
        pred_min = min(clf.predict_proba(X)[:,1])
        cut_range = np.linspace(int(pred_min * 100), int(pred_max * 100), int(pred_max * 100) - int(pred_min * 100) + 1) / 100
        PRTF_table = pd.DataFrame(list(map(_PRTF,cut_range)), columns=['Cut point','Precision','Recall(TPR)','FPR','F1-score'])

        fig, ax = plt.subplots()
        ax.plot(PRTF_table['Cut point'], PRTF_table['Precision'],'--', label='Precision Curve')
        ax.plot(PRTF_table['Cut point'], PRTF_table['Recall(TPR)'],'--', label='Recall Curve')
        ax.plot(PRTF_table['Cut point'], PRTF_table['F1-score'],'k', label='F1-score Curve')

        max_f1 = max(PRTF_table['F1-score'])
        best_cut = np.array(PRTF_table.loc[PRTF_table['F1-score'] == max_f1, 'Cut point']).mean()

        ax.plot([best_cut, best_cut], [max_f1 - 0.15, max_f1 + 0.05], color='red')
        ax.set_title('Precision, Recall, and F1-Score curves.')
        plt.text(best_cut, max_f1 - 0.2, 'Max F1:'+str(round(max_f1, 2))+' Cut point:'+str(best_cut))
        ax.legend()
        plt.show()

        return PRTF_table
    
    if 'cut' in calc_list:
        prtf = PRTF(X = X_train, y = y_train, clf = clf)
        print('>>>>>Best cut point on MAX(F1-score):\n', prtf.loc[prtf['F1-score'] == max(prtf['F1-score'])])
        
        if len(out_path):
            prtf.to_csv(out_path+'_cut.csv',index=False)
    
    # 模型Score 
    print('>>>>>Train set score:{:.2%}'.format(clf.score(X_train, y_train)))
    if len(X_test):
        print('>>>>>Test set score:{:.2%}'.format(clf.score(X_test, y_test)))
    
    # 4.2 AUC、ROC曲线、KS图
    # scikitplot系列图   
    # 该函数图完整引用ref7内容并轻微整理
    def roc_fig(**kwargs):    
        fpr, tpr, thresholds = roc_curve(y_train, clf.predict_proba(X_train)[:,1], pos_label = 1)
        auc_score = auc(fpr, tpr)
        
        if len(X_test):
            fpr1, tpr1, thresholds1 = roc_curve(y_test, clf.predict_proba(X_test)[:,1], pos_label = 1)
            auc_score1 = auc(fpr1, tpr1)
        
        fig, ax = plt.subplots()
        ax.plot(fpr, tpr, label='Train set AUC = %.5f'% auc_score)
        if len(X_test):
            ax.plot(fpr1, tpr1, label='Test set AUC = %.5f'% auc_score1)
        ax.set_title('Receiver Operating Characteristic')
        ax.plot([0, 1], [0, 1], '--', color = (0.6, 0.6, 0.6))  # 中轴线
        ax.legend()
        plt.show()
    
    if ('roc' in calc_list) or ('auc' in calc_list):
        roc_fig()
    
    # PR图（准召图）
    if 'pr' in calc_list:
        skplt.metrics.plot_precision_recall(y_test if len(y_test) else y_train, 
                                            clf.predict_proba(X_test if len(X_test) else X_train),
                                            classes_to_plot = [1], plot_micro = False)
        plt.show()
        if len(y_test) == 0:
            print('>>>>> Notice:This PR curve from train set.')
        
    # 学习速率图
    if 'lc' in calc_list:
        skplt.estimators.plot_learning_curve(clf, X_train, y_train)
        plt.show()
    
    # ks图
    if 'ks' in calc_list:
        skplt.metrics.plot_ks_statistic(y_test if len(y_test) else y_train, clf.predict_proba(X_test if len(X_test) else X_train))
        plt.show()
        if len(y_test) == 0:
            print('>>>>> Notice:This KS curve from train set.')
        
    # 特征重要性图 
#    hasattr(clf, 'feature_importances_')
    if 'fi' in calc_list:
        try:
            skplt.estimators.plot_feature_importances(clf, x_tick_rotation = 90, feature_names=X_train.columns)
            plt.show()
        
            if len(out_path):
                result_save = pd.ExcelWriter(out_path+'.xlsx')
                pd.DataFrame(list(zip(X_train.columns, clf.feature_importances_)), columns=['Var','imp']).to_excel(result_save,'Feature importance',encoding='utf-8',index=False)
                if 'cut' in calc_list:
                    prtf.to_excel(result_save,'Cut point',encoding='utf-8',index=False)
                result_save.save()
                
                pd.DataFrame(list(zip(X_train.columns, clf.feature_importances_)), columns=['Var','imp']).to_csv(out_path+'_importance.csv',index=False)
        except:
            pass
    
    # Calibration Curve
    if len(calibration):
        skplt.metrics.plot_calibration_curve(y_test, 
                                             list(map(lambda x:x.predict_proba(X_test),calibration.values())),
                                             list(calibration))
        plt.show()
        
    # ref7：https://www.cnblogs.com/gasongjian/p/8159501.html
        


# -*- coding: utf-8 -*-
# Author: tigflanker
# Date: 13 Dec 2018
# 无监督分箱（等频、等宽）

# 导包/宏参设定
import numpy as np
import pandas as pd 
from pandas._libs import lib
import math

def Cut_Merge(
        datain  # 输入主参数，可输入DataFrame或Series
        ,cut_way = 'ef'  # 等频: ef; 等宽: ew
        ,cut_y = ''  # 定义Y变量，用于分箱屏蔽
        ,cut_n = 3  # 分箱数（等宽模式下如果出现空箱，则逐步缩箱）
        ,calc_woe = True  # 连带计算WoE-IV，数据集直接以WoE转化后输出
        ):
        
    '''
    # Example:
    datain_example = pd.read_csv('http://biostat.mc.vanderbilt.edu/wiki/pub/Main/DataSets/titanic.txt')
    
    cut_df, fa_df = Cut_Merge(datain_example, cut_way = 'ew', cut_n = 10, cut_y = 'survived')
        
    # Example2:
    from sklearn.datasets import load_breast_cancer
    datain_example = load_breast_cancer()
    
    data_part = pd.DataFrame(datain_example.data, columns=datain_example.feature_names)
    flag_part = pd.Series(datain_example.target, name='y_var')
    
    datain_example = pd.concat([flag_part, data_part], axis=1)
    cut_df, fa_df = Cut_Merge(datain_example, cut_way = 'ef', cut_n = 10, cut_y = 'y_var')
    '''
    
    # 必须重置index
    datain = datain.reset_index(drop = True)
        
    # 主执行宏（数值型） 
    def _cut_merge(cut_var, cut_n = cut_n):       
        if cut_way == 'ew':
            cut_mass = pd.cut(np.array(cut_var), [-np.inf] + list(np.linspace(cut_var.min(), cut_var.max(), cut_n + 1))[1:cut_n] + [np.inf])
            divide_rule = cut_mass.categories
            
        elif cut_way == 'ef':
            divide_rule = [cut_var.quantile(x) for x in np.linspace(0, cut_n, cut_n + 1) / cut_n][1:cut_n]
            cut_mass = pd.cut(np.array(cut_var), [-np.inf] + sorted(set(divide_rule),key = divide_rule.index) + [np.inf])
            divide_rule = cut_mass.categories
        
        # 如果使用等宽分箱，避免出现空组
        while (cut_way == 'ew') & (len(set(cut_mass.codes) - {-1}) < cut_n - 1):
            cut_n -= 1
            cut_mass = pd.cut(np.array(cut_var), [-np.inf] + list(np.linspace(cut_var.min(), cut_var.max(), cut_n - 1)) + [np.inf])
            divide_rule = cut_mass.categories
        
        return pd.Series(cut_mass.codes, name=cut_var.name), {cut_var.name: divide_rule}
    
    # 主参判断 
    if type(datain) is pd.Series:
        cut_df, cut_divide_rule = _cut_merge(datain)
    elif type(datain) is pd.DataFrame:
        
        # 备份原数 
        cut_df = datain.copy()
        
        # 数据更新 
        print('Start cutting numerical variables...')
        _datatp = datain.dtypes    
        _map_result = list(map(lambda col:_cut_merge(datain[col]), 
                               set(_datatp[(_datatp != 'object') & (_datatp.index != cut_y)].index)))
        cut_df.update(pd.DataFrame(list(map(lambda u:u[0], _map_result))).T)
        
        # 字典整理 
        cut_divide_rule = {}
        for x in list(map(lambda u:u[1],_map_result)):
            cut_divide_rule.update(x)
        print('Numerical variables clear!')

        ## 字符型处理
        if sum(_datatp == 'object') > 0:
            print('Start cutting character variables...')
            cut_df_char = pd.DataFrame(index = cut_df.index)
            for _char_var in list(_datatp[(_datatp == 'object') & (_datatp.index != cut_y)].index):
                _sub_dict = {}
                for i, value in enumerate(sorted(list(set(cut_df[_char_var])),key=lambda x:x is not np.nan)):
                    _sub_dict[value] = i if np.nan not in set(cut_df[_char_var]) else i - 1

                # 字典更新
                cut_divide_rule.update({_char_var:list(_sub_dict)})
                
                # 值更新
                cut_df_char = pd.concat([cut_df_char, pd.Series(list(map(lambda y:_sub_dict[y], cut_df[_char_var])), name=_char_var)], axis=1)
            
            cut_df = pd.concat([cut_df[_datatp[(_datatp != 'object')].index], cut_df_char], axis=1)      
#            cut_df = pd.merge(cut_df[_datatp[(_datatp != 'object')].index], cut_df_char, left_index=True, right_index=True)
            print('Character variables clear!')
            
    else:
        print('Only pd.DataFrame or pd.Series data type supported in this version.')
        cut_df, cut_divide_rule = 0, 0
        
    # Feature Analysis Dataframe
    if type(datain) is pd.DataFrame:
        # 箱内情况统计
        def _tmp_func(var):
            _agged = cut_df.groupby(var)[cut_y].agg([sum,'count',np.mean]) 
            _agged['Var'], _agged['Binn'] = var, _agged.index
            
            return _agged
        
        rate_block = pd.concat(list(map(_tmp_func, cut_divide_rule)))
        rate_block = rate_block.rename(index=str, columns={'sum':'bads','count':'N','mean':'bad rate'})
        rate_block['bad rate'] = ['%.2f%%' % (i * 100) for i in rate_block['bad rate']]
            
        # 框架 
        fa_df = pd.concat(list(map(lambda x:pd.Series(cut_divide_rule[x], index=[x] * len(cut_divide_rule[x])), list(cut_divide_rule))))    
        fa_df = pd.DataFrame({'Var':list(fa_df.index),'Bins':list(fa_df)})
        fa_df = fa_df.dropna()
        
        fa_df['Binn'] = fa_df.index
        fa_df['Binn'] = fa_df.Binn.groupby(fa_df.Var).rank(method='first').apply(lambda x:int(x - 1))
        
        # dummy出nan情况 
        if len(rate_block.loc[rate_block.Binn == -1,'Var']):
            _dummy_nan = pd.DataFrame(rate_block.loc[rate_block.Binn == -1,'Var'])
            _dummy_nan['Bins'], _dummy_nan['Binn'] = np.nan, -1
            fa_df = pd.concat([fa_df, _dummy_nan], axis=0, ignore_index=True)
        
        fa_df = pd.merge(fa_df, rate_block, how = 'right', on = ['Var', 'Binn'])
        fa_df = fa_df.sort_values(['Var', 'Binn']).reset_index(drop=True)
    else:
        fa_df = []
        
    # 是否连带计算WoE-IV情况，并将结果以WoE转化后输出
    if calc_woe:
        print('Calculating WoE & IV value...')
        print('!!!Notice: The value in output dataset will be transform into WoE.!!!')
        
        # 计算好坏总数
        _sum_set = fa_df.groupby('Var').agg({'bads':sum, 'N':sum}).rename(columns={'bads':'bads_sum','N':'N_sum'})
        fa_df = pd.merge(fa_df, _sum_set, how='left', left_on='Var', right_index=True)
        
        # WoE_IV计算主方法 
        def _WoE_IV(x):
            LS = int(x.bads in (0, x.N))
            
            PYi = (x.bads + LS) / (x.bads_sum + LS)
            PNi = (x.N - x.bads + LS) / (x.N_sum - x.bads_sum + LS)
                
            WoE = math.log(PYi / PNi)   
            return WoE, (PYi - PNi) * WoE
           
        fa_df = pd.concat([fa_df, pd.DataFrame(list(fa_df.apply(_WoE_IV, axis = 1)), columns=['Woe', 'Iv'])], axis=1) 
        _iv_set = pd.DataFrame(fa_df.groupby('Var').Iv.sum(), columns = ['Iv'])
        _iv_set = pd.concat([_iv_set, _iv_set.Iv.rank(ascending=False).rename('Iv_rank')], axis=1)
        fa_df = pd.merge(fa_df.drop(['Iv', 'bads_sum', 'N_sum'], axis=1), _iv_set, how='left', left_on='Var', right_index=True)    
        
        # 2.3.1.3 WOE数据集映射 
        wi_dataout = cut_df.copy()
        
        woe_dict = {}
        for col in set(fa_df.Var):
            woe_dict[col] = dict(zip(fa_df.Binn[fa_df.Var == col], fa_df.Woe[fa_df.Var == col]))
        
    #         Way 1:
    #         cut_df.update(wi_dataout[list(set(fa_df.Var))].apply(lambda x:x.map(woe_dict[x.name])))
    #         wi_dataout = wi_dataout[list(set(fa_df.Var))].apply(lambda x:x.replace(woe_dict[x.name]))
    
    #         Way 2:
    #         wi_dataout = []
    #         for col, value in cut_df[list(set(fa_df.Var))].iteritems():
    #             wi_dataout.append(value.map(woe_dict[col]))
    #         wi_dataout = pd.DataFrame(wi_dataout).T
            
    #         Way 3:
        def dictmap(self, dictx):
            def infer(x):
                dictx_sub = dictx[x.name]
                if x.empty:
                    return lib.map_infer(x, lambda d:dictx_sub.get(d))
                return lib.map_infer(x.astype(object).values, lambda d:dictx_sub.get(d))
        
            return self.apply(infer)
        
        wi_dataout = dictmap(cut_df[list(set(fa_df.Var))], woe_dict)
        
        # final erge
        print('Final merge step...')
        cut_df = pd.merge(cut_df.drop(list(wi_dataout.columns), axis = 1), wi_dataout, left_index = True, right_index = True)

    return cut_df, fa_df
    


# -*- coding: utf-8 -*-
# Author: tigflanker
# Date: 17 Dec 2018
# 哑变量处理 

# 导包/宏参设定
import pandas as pd 

def Dummy(datain
          ,dummy_var = []  # 待处理变量，接受单str或list、set可循环内容 
          ,drop_one = True  # 是否将首值"第0个映射"删除，以防模型特征共线性 
          ,drop_orig = True  # 是否删除数据集原始变量 
          ):
          
    '''
    # Example:
    datain_test = pd.read_csv('http://biostat.mc.vanderbilt.edu/wiki/pub/Main/DataSets/titanic.txt')
    
    # 单序列dummy
    dummy_mapping_df, dummy_mapping_dict = Dummy(datain_test, dummy_var = 'pclass')  
    
    # 数据集dummy
    dummy_mapping_df, dummy_mapping_dict = Dummy(datain_test[['pclass','survived','age','sex']])  # dummy全集
    dummy_mapping_df, dummy_mapping_dict = Dummy(datain_test, dummy_var = ['pclass','survived','age','sex'])  # dummy指定特征（字符数值均可）
    
    # 参数测试
    dummy_mapping_df, dummy_mapping_dict = Dummy(datain_test, drop_orig=False, drop_one=False)
    '''
    
    # 参数解析 
    if type(dummy_var) == str:
        dummy_var = [dummy_var]
    elif len(dummy_var) == 0:
        dummy_var = list(datain.dtypes[datain.dtypes == 'object'].index)
        
    # dummy主程序 
    def _dummy_temp(x):
        dummy_value = [0] * len(dummy_value_list)
        if x in dummy_value_list:
            dummy_value[dummy_value_list.index(x)] = 1
        
        return dummy_value    
    
    # 生成参数数据和映射字典 
    dummy_mapping_df = datain.copy()
    dummy_mapping_dict = {}
    for i, dv in enumerate(dummy_var):
        dummy_value_list = list(datain[dv].unique())
        dummy_mapping_dict[dv] = [[dv+'_'+str(x), dummy_value_list[x]] for x in range(len(dummy_value_list))]
        dummy_mapping_df = pd.concat([dummy_mapping_df,pd.DataFrame(list(map(_dummy_temp, datain[dv]))
            , columns=[dv+'_'+str(x) for x in range(len(dummy_value_list))])], axis=1)
    
        if drop_orig:
            del dummy_mapping_df[dv]
        if drop_one:
            del dummy_mapping_df[dv+'_0']
            dummy_mapping_dict[dv] = dummy_mapping_dict[dv][1:]
                    
    return dummy_mapping_df, dummy_mapping_dict
